///////////////////////////////////////////////////////////////////////////////
// Project:     M - cross platform e-mail GUI client
// File name:   util/ssl.cpp
// Purpose:     SSL support functions
// Author:      Karsten Ballder
// Modified by:
// Created:     24.01.01 (extracted from MailFolderCC.cpp)
// CVS-ID:      $Id: ssl.cpp,v 1.23 2002/01/12 23:19:37 vadz Exp $
// Copyright:   (C) 1997-2000 by Karsten Ballder (ballueder@gmx.net)
// Licence:     M license
///////////////////////////////////////////////////////////////////////////////

#include "Mpch.h"

#ifndef USE_PCH
   #include "Mcommon.h"
#endif // !USE_PCH

#ifdef USE_SSL

#ifndef USE_PCH
   #include "Profile.h"
   #include "MApplication.h"
#endif // !USE_PCH

#include "Mdefaults.h"

#include "MDialogs.h"

#include <wx/dynlib.h>

extern const MOption MP_SSL_DLL_CRYPTO;
extern const MOption MP_SSL_DLL_SSL;

/* These functions will serve as stubs for the real openSSL library
   and must be initialised at runtime as c-client actually links
   against these. */

#include <openssl/ssl.h>

// starting from 0.9.6a, OpenSSL uses void, as it should, instead of char
#include <openssl/opensslv.h>
#if defined(OPENSSL_VERSION_NUMBER) && (OPENSSL_VERSION_NUMBER > 0x0090600fL)
   #define ssl_data_t void *
#else // old ssl
   #define ssl_data_t char *
#endif

/* This is our interface to the library and auth_ssl.c in c-client
   which are all in "C" */
extern "C" {

// defined in cclient auth_ssl.c (but not declared in any public header)
extern void ssl_onceonlyinit();

#define SSL_DEF(returnval, name, args) \
   typedef returnval (* name##_TYPE) args ; \
   name##_TYPE stub_##name = NULL

#define SSL_LOOKUP(name) \
      stub_##name = (name##_TYPE) gs_dllSll.GetSymbol(#name); \
      if ( !stub_##name ) \
         goto error

#define CRYPTO_LOOKUP(name) \
      stub_##name = (name##_TYPE) gs_dllCrypto.GetSymbol(#name); \
      if ( !stub_##name ) \
         goto error

SSL_DEF( SSL *, SSL_new, (SSL_CTX *ctx) );
SSL_DEF( void, SSL_free, (SSL *ssl) );
SSL_DEF( int,  SSL_set_rfd, (SSL *s, int fd) );
SSL_DEF( int,  SSL_set_wfd, (SSL *s, int fd) );
SSL_DEF( void, SSL_set_read_ahead, (SSL *s, int yes) );
SSL_DEF( int,  SSL_connect, (SSL *ssl) );
SSL_DEF( int,  SSL_read, (SSL *ssl,ssl_data_t buf,int num) );
SSL_DEF( int,  SSL_write, (SSL *ssl,const ssl_data_t buf,int num) );
SSL_DEF( int,  SSL_pending, (SSL *s) );
SSL_DEF( int,  SSL_library_init, (void ) );
SSL_DEF( void, SSL_load_error_strings, (void ) );
SSL_DEF( SSL_CTX *,SSL_CTX_new, (SSL_METHOD *meth) );
SSL_DEF( unsigned long, ERR_get_error, (void) );
SSL_DEF( char *, ERR_error_string, (unsigned long e, char *p));
//extern SSL_get_cipher_bits();
SSL_DEF( const char *, SSL_CIPHER_get_name, (SSL_CIPHER *c) );
SSL_DEF( int, SSL_CIPHER_get_bits, (SSL_CIPHER *c, int *alg_bits) );
//extern SSL_get_cipher();
SSL_DEF( SSL_CIPHER *, SSL_get_current_cipher ,(SSL *s) );
#   if defined(SSLV3ONLYSERVER) && !defined(TLSV1ONLYSERVER)
SSL_DEF(SSL_METHOD *, SSLv3_client_method, (void) );
#   elif defined(TLSV1ONLYSERVER) && !defined(SSLV3ONLYSERVER)
SSL_DEF(int, TLSv1_client_method, () );
#   else
SSL_DEF(SSL_METHOD *, SSLv23_client_method, (void) );
#   endif

#undef SSL_DEF

SSL     * SSL_new(SSL_CTX *ctx)
{ return (*stub_SSL_new)(ctx); }
void  SSL_free(SSL *ssl)
{ (*stub_SSL_free)(ssl); }
int  SSL_set_rfd(SSL *s, int fd)
{ return (*stub_SSL_set_rfd)(s,fd); }
int  SSL_set_wfd(SSL *s, int fd)
{ return (*stub_SSL_set_wfd)(s,fd); }
void  SSL_set_read_ahead(SSL *s, int yes)
{ (*stub_SSL_set_read_ahead)(s,yes); }
int   SSL_connect(SSL *ssl)
{ return (*stub_SSL_connect)(ssl); }
int   SSL_read(SSL *ssl,ssl_data_t buf, int num)
{ return (*stub_SSL_read)(ssl, buf, num); }
int   SSL_write(SSL *ssl,const ssl_data_t buf,int num)
{ return (*stub_SSL_write)(ssl, buf, num); }
int  SSL_pending(SSL *s)
{ return (*stub_SSL_pending)(s); }
int       SSL_library_init(void )
{ return (*stub_SSL_library_init)(); }
void  SSL_load_error_strings(void )
{ (*stub_SSL_load_error_strings)(); }
SSL_CTX * SSL_CTX_new(SSL_METHOD *meth)
{ return (*stub_SSL_CTX_new)(meth); }
unsigned long ERR_get_error(void)
{ return (*stub_ERR_get_error)(); }
char * ERR_error_string(unsigned long e, char *p)
{ return (*stub_ERR_error_string)(e,p); }
const char * SSL_CIPHER_get_name(SSL_CIPHER *c)
{ return (*stub_SSL_CIPHER_get_name)(c); }
SSL_CIPHER * SSL_get_current_cipher(SSL *s)
{ return (*stub_SSL_get_current_cipher)(s); }
int SSL_CIPHER_get_bits(SSL_CIPHER *c, int *alg_bits)
{
  return (*stub_SSL_CIPHER_get_bits)(c,alg_bits);
}


#   if defined(SSLV3ONLYSERVER) && !defined(TLSV1ONLYSERVER)
SSL_METHOD *  SSLv3_client_method(void)
{ return (*stub_SSLv3_client_method)(); }
#   elif defined(TLSV1ONLYSERVER) && !defined(SSLV3ONLYSERVER)
int TLSv1_client_method(void)
{ (* stub_TLSv1_client_method)(); }
#   else
SSL_METHOD * SSLv23_client_method(void)
{ return (* stub_SSLv23_client_method)(); }
#   endif

} // extern "C"

static bool gs_SSL_loaded = false;
static bool gs_SSL_available = false;

static wxDynamicLibrary gs_dllSll,
                        gs_dllCrypto;

bool InitSSL(void) /* FIXME: MT */
{
   static bool s_errMsgGiven = false;

   if(gs_SSL_loaded)
      return gs_SSL_available;

   String ssl_dll = READ_APPCONFIG(MP_SSL_DLL_SSL);
   String crypto_dll = READ_APPCONFIG(MP_SSL_DLL_CRYPTO);

   // it doesn't take long so nobody sees this message anyhow
#if 0
   STATUSMESSAGE((_("Trying to load SSL libraries '%s' and '%s'..."),
                  crypto_dll.c_str(),
                  ssl_dll.c_str()));
#endif // 0

   if ( !gs_dllCrypto.Load(crypto_dll) )
      goto error;

   if ( !gs_dllSll.Load(ssl_dll) )
      goto error;

   SSL_LOOKUP(SSL_new );
   SSL_LOOKUP(SSL_free );
   SSL_LOOKUP(SSL_set_rfd );
   SSL_LOOKUP(SSL_set_wfd );
   SSL_LOOKUP(SSL_set_read_ahead );
   SSL_LOOKUP(SSL_connect );
   SSL_LOOKUP(SSL_read );
   SSL_LOOKUP(SSL_write );
   SSL_LOOKUP(SSL_pending );
   SSL_LOOKUP(SSL_library_init );
   SSL_LOOKUP(SSL_load_error_strings );
   SSL_LOOKUP(SSL_CTX_new );
   SSL_LOOKUP(SSL_CIPHER_get_name );
   SSL_LOOKUP(SSL_get_current_cipher );
   CRYPTO_LOOKUP(ERR_get_error);
   CRYPTO_LOOKUP(ERR_error_string);
   SSL_LOOKUP(SSL_CIPHER_get_bits);
#if defined(SSLV3ONLYSERVER) && !defined(TLSV1ONLYSERVER)
   SSL_LOOKUP(SSLv3_client_method );
#elif defined(TLSV1ONLYSERVER) && !defined(SSLV3ONLYSERVER)
   SSL_LOOKUP(TLSv1_client_method );
#else
   SSL_LOOKUP(SSLv23_client_method );
#endif

   gs_SSL_available =
   gs_SSL_loaded = true;

   STATUSMESSAGE((_("Successfully loaded '%s' and '%s' - "
                    "SSL authentification is now available."),
                  crypto_dll.c_str(),
                  ssl_dll.c_str()));

   ssl_onceonlyinit();

   return true;

error:
   if ( !s_errMsgGiven )
   {
      ERRORMESSAGE((_("SSL authentication is not available.")));

      s_errMsgGiven = true;

      // show the log dialog first
      wxLog::FlushActive();

      MDialog_Message
      (
         _("You can change the locations of the SSL and crypto "
           "libraries in the last page of the preferences dialog\n"
           "if you have these libraries in non default location"
           " or if they have some other names on your system."),
         NULL,
         "SSL tip",
         "SSLLibTip"
      );
   }

   return false;
}

#undef SSL_LOOKUP

#else // !USE_SSL

bool InitSSL(void)
{
   static bool s_errMsgGiven = false;
   if ( !s_errMsgGiven )
   {
      s_errMsgGiven = true;

      ERRORMESSAGE((_("This version of the program doesn't support SSL "
                      "authentification.")));
   }

   return false;
}

#endif // USE_SSL/!USE_SSL


